// Copyright 2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "JuliaView.h"

#import <AppKit/NSBitmapImageRep.h>
#import <CoreGraphics/CGContext.h>
#import <stdlib.h>

#define WIDTH          (3.5)

#define INTEREST_MAP_WIDTH  300
#define INTEREST_MAP_HEIGHT 300


static void _JuliaWorkerProcessWorkBuffer(Worker *worker);


#define CHANGE_MODES

unsigned int OFNumberOfProcessors();

@interface JuliaView (PrivateAPI)
- (void) _randomJump;
- (void) _startBuffer: (JuliaWorkBuffer *) worker;
- (void)_internalStep;
@end


@implementation JuliaView

- initWithFrame: (NSRect) rect;
{    
    if (!(self = [super initWithFrame: rect]))
        return nil;

    testPoint.orbit = NSZoneMalloc(NULL, sizeof(float[4]) * MAX_ITERATIONS);
    testPoint.N = MAX_ITERATIONS;
    
    return self;
}

- (void) dealloc;
{
    NSZoneFree(NULL, testPoint.orbit);

    if (workers) {
        while (workerCount--)
            NSZoneFree(NULL, workers[workerIndex].image);
        NSZoneFree(NULL, workers);
    }
    [super dealloc];
}

- (BOOL) isStandalone;
{
    // Under MacOS X Public Beta, the screen saver gets displayed in a tiny window
    // in the System Preferences app when you are selecting the screen saver to use.
    return [[self superview] superview] == nil;
}

- (void) _oneStep: (id) arg;
{
    if (![_window isVisible])
        return;
    [super _oneStep: arg];
}

- (void) oneStep;
{
    // The animation timer is too slow.
    unsigned int i;
    
    for (i = 0; i < 100; i++) {
        [self _internalStep];
    }
}

- (NSTimeInterval)animationDelayTime;
{
    // We don't need any help being slow, but if this is too low, the exciting
    // stuff (which is easier to compute that the boring stuff) goes by too quick
    return 0.0;
}

- (BOOL)useBufferedWindow;
{
    return NO;
}

- (BOOL)useMultipleMonitors; // FreeSpace Only
{
    // We are slow enough on one monitor
    return NO;
}

#define MAX_DISPLAYS 128

- (void)startAnimation;
{
    CGDisplayErr err;
    NSArray *displayModes;
    NSDictionary *mode = nil;
    unsigned int modeCount, modeIndex;
    BOOL foundMode = NO;
    NSRect windowRect;
    BOOL verbose = NO;
    unsigned int pixelIndex;
    unsigned int i;
    unsigned int desiredWidth = 640;
    unsigned int desiredHeight = 480;

    // Record the current mode and description of that mode
    originalMode = [(NSDictionary *)CGDisplayCurrentMode(kCGDirectMainDisplay) retain];

    // Look for a 32-bit mode of the desired size
    displayModes = (NSArray *)CGDisplayAvailableModes(kCGDirectMainDisplay);
    if (!displayModes) {
        [NSException raise: NSInternalInconsistencyException
                    format: @"CGDisplayAvailableModes returned NULL"];
    }

    modeCount = [displayModes count];
    if (verbose) {
        NSLog(@"%d modes avaliable", modeCount);
        NSLog(@"Current mode is %s", [[originalMode description] cString]);
    }


    for (modeIndex = 0; modeIndex < modeCount; modeIndex++) {
        mode = [displayModes objectAtIndex: modeIndex];
        if (verbose)
            NSLog(@" mode %d -- %s", modeIndex, [[mode description] cString]);

        if ([[mode objectForKey: (id)kCGDisplayWidth] intValue] != desiredWidth ||
            [[mode objectForKey: (id)kCGDisplayHeight] intValue] != desiredHeight) {
            if (verbose)
                NSLog(@" -- bad size");
            continue;
        }

        if ([[mode objectForKey: (id)kCGDisplayBitsPerPixel] intValue] != 32) {
            if (verbose)
                NSLog(@" -- bad depth");
            continue;
        }

        // Found a good mode!
        foundMode = YES;
        break;
    }

    if (!foundMode)
        [NSException raise: NSInternalInconsistencyException
                    format: @"Couldn't find a %dx%dx32 display mode.", desiredWidth, desiredHeight];

    saverMode = [mode retain];
    if (verbose)
        NSLog(@" saverMode -- %s", [[saverMode description] cString]);

#ifdef CHANGE_MODES
    if ([self isStandalone]) {
        // Lock the display to make the mode change synchronous and to avoid notifying other apps
        // of the screen size notification (we don't want them to resize their windows).
        if (!CGDisplayIsCaptured(kCGDirectMainDisplay)) {
            err = CGDisplayCapture(kCGDirectMainDisplay);
            if (err != CGDisplayNoErr) {
                [NSException raise: NSInternalInconsistencyException
                            format: @"Unable to capture display"];
            }
        }
        err = CGDisplaySwitchToMode(kCGDirectMainDisplay, (CFDictionaryRef)saverMode);
        if ( err != CGDisplayNoErr ) {
            [NSException raise: NSInternalInconsistencyException
                        format: @"Unable to set display mode"];
        }
    }
#endif
    

    // Reposition our window appropraitely.  The correct position is the upper-left corner
    // of the old display area (with zero,zero being the lower left).  The AppKit does NOT
    // know about display mode changes yet, so we can't count on NSScreen to have the
    // right size -- we'll get all the size info from the display mode description structs.
    // This window is never reused, so we don't need to set the frame back in
    // -willExitScreenSaverMode.
    if ([self isStandalone]) {
        windowRect = NSMakeRect(0,
                                [[originalMode objectForKey: (id)kCGDisplayHeight] intValue] -
                                [[saverMode objectForKey: (id)kCGDisplayHeight] intValue],
                                [[saverMode objectForKey: (id)kCGDisplayWidth] intValue],
                                [[saverMode objectForKey: (id)kCGDisplayHeight] intValue]);

/*
        windowRect = NSMakeRect(0,
                                0,
                                [[originalMode objectForKey: (id)kCGDisplayWidth] intValue],
                                [[originalMode objectForKey: (id)kCGDisplayHeight] intValue]);
        NSLog(@"setting windowRect = %@", NSStringFromRect(windowRect));
*/

        [[self window] setFrame: windowRect display: NO];
        
        [self lockFocus];
        [[NSColor blueColor] set];
        NSRectFill(_frame);
        [self unlockFocus];
        
        [[self window] setLevel: CGShieldingWindowLevel()];
    }
    
    pixelWidth = _frame.size.width;
    pixelHeight = _frame.size.height;

    pixelWidth -= pixelWidth % 4;
    pixelHeight -= pixelHeight % 2;
    
    srandom(time(NULL));
    
    // We'll always start at the zero point right now (which is a circle).
    c1[0] = c1[1] = c1[2] = c1[3] = 0.0;
    currentAngle = 0.0;
    t = 0.0;
    [self _randomJump];

    workerCount = OFNumberOfProcessors();
    workerIndex = 0;
    workers = NSZoneCalloc(NULL, workerCount, sizeof(*workers));
    
    for (i = 0; i < workerCount; i++) {
        JuliaWorkBuffer *worker;
        color_t *image;
        
        worker = &workers[i];
        WorkerInit(&worker->worker, _JuliaWorkerProcessWorkBuffer);
        
        worker->pixelWidth = pixelWidth;
        worker->pixelHeight = pixelHeight;
        
        worker->image = image = NSZoneMalloc(NULL, pixelWidth * pixelHeight * sizeof(*image));    
        for (pixelIndex = 0; pixelIndex < pixelWidth * pixelHeight; pixelIndex++) {
            image[pixelIndex].r = 0;
            image[pixelIndex].g = 0;
            image[pixelIndex].b = 0;
            image[pixelIndex].a = 255;
        }
        
        // Record all the information about the size of the view
        worker->width = WIDTH;
        worker->height = ((float)WIDTH/(float)pixelWidth) * pixelHeight;

        worker->center[0] = 0.0;
        worker->center[1] = 0.0;
        worker->center[2] = 0.0;
        worker->center[3] = 0.0;
    
        worker->dx[0] = WIDTH / (float)pixelWidth;
        worker->dx[1] = 0.0;
        worker->dx[2] = 0.0;
        worker->dx[3] = 0.0;
    
        worker->dy[0] = 0.0;
        worker->dy[1] = worker->height / (float)pixelHeight;
        worker->dy[2] = 0.0;
        worker->dy[3] = 0.0;
        
        worker->clutIndex = 0;

        // Start the workers and give them all initial buffers to work on
        WorkerStartBackgroundThread(&worker->worker);
        [self _startBuffer: worker];
    }

    [NSCursor setHiddenUntilMouseMoves: YES];
    [super startAnimation];
}

- (void)stopAnimation;
{
    CGDisplayErr err;
    unsigned int i;
    
    
    // Stop all the running workers
    for (i = 0; i < workerCount; i++) {
        JuliaWorkBuffer *worker;
        
        worker = &workers[i];
        WorkerDestroy(&worker->worker);
        
        NSZoneFree(NULL, worker->image);
    }
    
    NSZoneFree(NULL, workers);
    workers = NULL;
    
    
    // Change back to the original screen resolution (if necessary)
#ifdef CHANGE_MODES
    if ([self isStandalone]) {
        err = CGDisplaySwitchToMode(kCGDirectMainDisplay, (CFDictionaryRef)originalMode);
        if ( err != CGDisplayNoErr ) {
            [NSException raise: NSInternalInconsistencyException
                        format: @"Unable to restore display mode"];
        }
    
        if (CGDisplayIsCaptured(kCGDirectMainDisplay)) {
            err = CGDisplayRelease(kCGDirectMainDisplay);
            if (err != CGDisplayNoErr) {
                [NSException raise: NSInternalInconsistencyException
                            format: @"Unable to release display"];
            }
        }
    }
#endif

    [super stopAnimation];
}

@end



// Want to return a number between 0.0 and 1.0
static inline float frandom()
{
    long l = random();

    return l / (float)LONG_MAX;
}


@implementation JuliaView (PrivateAPI)


#define ANGLE_STEP       (2*M_PI / 31.0)
#define RAY_LENGTH_START (4.0)
#define DESIRED_DISTANCE (0.001)
#define RANDOM_OFFSET    (0.2)

// We'll generate points to visit by building a ray from the origin (0,0) at an angle.
// Each time we pick a point, this angle will increase.  Once we have the ray, we'll
// find a point along the ray that is within some cutoff distance of the set.  We
// will never pick points that are inside the set as end points here, but as we are
// animating between the end points, we will hits lots of interior points.
- (void) _randomJump;
{
    float rayEndX, rayEndY, rayLength, newLength;
    float offset;
    
    newLength = RAY_LENGTH_START;
    rayEndX = cos(currentAngle);
    rayEndY = sin(currentAngle);
    
    c0[0] = c1[0];
    c0[1] = c1[1];
    c0[2] = c1[2];
    c0[3] = c1[3];
    
    do {
        rayLength = newLength;
        testPoint.c[0] = rayLength * rayEndX;
        testPoint.c[1] = rayLength * rayEndY;
        testPoint.c[2] = 0.0;
        testPoint.c[3] = 0.0;
        
        qJulia(&testPoint);
                
        newLength -= 0.5 * testPoint.mandelbrotDistanceEstimate;
    } while (newLength > 0.0 && testPoint.mandelbrotDistanceEstimate > DESIRED_DISTANCE);
    
//    fprintf(stderr, "angle = %f, length = %f, c = (%f, %f), type = %d, dist = %f\n",
//            currentAngle, rayLength, testPoint.c[0], testPoint.c[1], testPoint.mandelbrotDistanceType,
//            testPoint.mandelbrotDistanceEstimate);

    // -1.0 .. 1.0
    offset = 2.0 * (frandom() - 0.5);
    
    offset *= RANDOM_OFFSET;
    rayLength += (offset * RANDOM_OFFSET);
    
    c1[0] = rayLength * rayEndX;
    c1[1] = rayLength * rayEndY;
    c1[2] = 0.0;
    c1[3] = 0.0;
    
    // Make this dependant upon the points/
    step = 1.0 / 60.0;
    
    currentAngle += ANGLE_STEP;
    while (currentAngle > 2*M_PI)
        currentAngle -= 2*M_PI;
}


- (void) _startBuffer: (JuliaWorkBuffer *) worker;
{
    if (t >= 1.0) {
        t = 0.0;
        [self _randomJump];
    }


    worker->c[0] = (1 - t) * c0[0] + t * c1[0];
    worker->c[1] = (1 - t) * c0[1] + t * c1[1];
    worker->c[2] = (1 - t) * c0[2] + t * c1[2];
    worker->c[3] = (1 - t) * c0[3] + t * c1[3];

    testPoint.c[0] = worker->c[0];
    testPoint.c[1] = worker->c[1];
    testPoint.c[2] = worker->c[2];
    testPoint.c[3] = worker->c[3];
    qJulia(&testPoint);
    worker->mandelbrotDistanceType = testPoint.mandelbrotDistanceType;

// TJW - This makes us step less (up to a point) when we are closer to the set.
// There are a couple problems with this:
//  1) We still end up jumping really quick when we are more than 1.0 from the set.
//     Sometimes the distance estimate isn't a good estimate on how 'dynamic' the
//     set is around the current point
//  2) Since we are only rendering at 640x480, we gets lots of sparklies in the fine
//     detail of the set as we move around and this looks yucky.
//    if (testPoint.mandelbrotDistanceType != InSetDistance &&
//        testPoint.mandelbrotDistanceEstimate > 0.0 &&
//        testPoint.mandelbrotDistanceEstimate < 1.0)
//        t += step * MAX(0.01, sqrt(testPoint.mandelbrotDistanceEstimate));
//    else
        t += step;

    WorkerSignalMoreWork(&worker->worker);
}

- (void)_internalStep;
{
    NSGraphicsContext *graphicsContext;
    CGContextRef ctx;
    JuliaWorkBuffer *worker;
    NSBitmapImageRep *imageRep;
    
    graphicsContext = [NSGraphicsContext currentContext];
    ctx = [graphicsContext graphicsPort];
    worker = workers + workerIndex;
    
    // Wait for this buffer to finish
    WorkerWaitForCompletion(&worker->worker);
    
    // Draw the results
    imageRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes: (unsigned char **)&worker->image
                                                       pixelsWide: pixelWidth
                                                       pixelsHigh: pixelHeight
                                                    bitsPerSample: 8
                                                  samplesPerPixel: 4
                                                         hasAlpha: YES
                                                         isPlanar: NO
                                                   colorSpaceName: NSDeviceRGBColorSpace
                                                      bytesPerRow: 0
                                                     bitsPerPixel: 0];
    [imageRep drawAtPoint: _frame.origin];
    [imageRep release];
    
    [self _startBuffer: worker];

    [_window flushWindow];
}


@end




//
//  Copied from OmniFoundation
//

#import <mach/mach.h>
#import <mach/mach_error.h>
#import <mach/bootstrap.h>

static BOOL         _OFMachInitialized = NO;
static host_priv_t  _OFHostPort = PORT_NULL;
static unsigned int _OFNumberOfProcessors = 0;

static void _OFInitHost()
{
    kern_return_t krc;
    processor_array_t processors;
    mach_port_t dummy_port_1, dummy_port_2, dummy_port_3, dummy_port_4;

    if (_OFMachInitialized)
        return;
    
    _OFMachInitialized = YES;
    
    krc = bootstrap_ports(mach_task_self(),
                          &_OFHostPort,
                          &dummy_port_1, &dummy_port_2,
                          &dummy_port_3, &dummy_port_4);
    if (krc != KERN_SUCCESS) {
        fprintf(stderr, "_OFInitHost: bootstrap_ports -- %s\n", mach_error_string(krc));
        return;
    }
    
    krc = host_processors(_OFHostPort, &processors, &_OFNumberOfProcessors);
    if (krc != KERN_SUCCESS) {
        fprintf(stderr, "_OFInitHost: host_processors -- %s\n", mach_error_string(krc));
        return;
    }
    
    // We don't actually need the list of processors...
    krc = vm_deallocate(mach_task_self(), (vm_address_t)processors, sizeof(*processors) * _OFNumberOfProcessors);
    if (krc != KERN_SUCCESS) {
        fprintf(stderr, "_OFInitHost: vm_deallocate -- %s\n", mach_error_string(krc));
        return;
    }
}

unsigned int OFNumberOfProcessors()
{
    if (!_OFMachInitialized)
        _OFInitHost();
    return _OFNumberOfProcessors;
}


// All the actual work is done here
static void _JuliaWorkerProcessWorkBuffer(Worker *worker)
{
    JuliaWorkBuffer *juliaWork = (JuliaWorkBuffer *)worker;
    
    //fprintf(stderr, "0x%08x: _JuliaWorkerProcessWorkBuffer\n", juliaWork);

    // We are guaranteed that the image is symmetric (x,y) == (-x,-y).
    // This means that we can calculate only half the number of pixels.  Care must
    // be taken, though, to not duplicate the 'zero' row.  Instead, the bottom row
    // on the top half should be shifted up one half a pixel width and thus the top
    // row on the bottom half would be correct.  This assumes that the pixel height
    // is even.  Were it odd for some reason, we would need to compute the top half,
    // flip that for the bottom half and compute the zero row on its own
    // (or compute N/2+1 pixels for the top and only flip N/2 (rounding down) pixels
    // to the bottom).  We'll assume an even height :)
    float topCenter[2];
    color_t *end, *start;
    unsigned int pixelCount;


    topCenter[0] = juliaWork->center[0] + (0.25 * juliaWork->pixelHeight + 0.5) * juliaWork->dy[0];
    topCenter[1] = juliaWork->center[1] + (0.25 * juliaWork->pixelHeight + 0.5) * juliaWork->dy[1];

    cComputeJuliaImage4(juliaWork->image,
                        topCenter, juliaWork->dx, juliaWork->dy,
                        juliaWork->pixelWidth,  juliaWork->pixelHeight / 2,
                        juliaWork->c, MAX_ITERATIONS,
                        clutTable[juliaWork->clutIndex],
                        juliaWork->mandelbrotDistanceType == InSetDistance,
                        0);
            
    // Now, copy the top part of the image to the bottom part, reversing
    // it as we go.  We could try to write some cool Altivec algorithm
    // for this, but in truth, this part will probably be a very small
    // percentage of the total time.
    start = juliaWork->image;
    end = juliaWork->image + (unsigned int)(juliaWork->pixelHeight * juliaWork->pixelWidth - 1);
    pixelCount = juliaWork->pixelHeight * juliaWork->pixelWidth / 2;
    while (pixelCount--) {
        *end = *start;
        end--;
        start++;
    }
}
