// Copyright 2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "Worker.h"

#import <stdio.h>


static void *_WorkerThread(void *arg);

void WorkerInit(Worker *worker, WorkerFunction function)
{
    //fprintf(stderr, "0x%08x: JuliaWorkerInit\n", worker);
    pthread_mutex_init(&worker->mutex, NULL);
    pthread_cond_init(&worker->condition, NULL);
    worker->thread = NULL;
    worker->state = WorkerStoppedState;
    worker->function = function;
}

void WorkerDestroy(Worker *worker)
{
    //fprintf(stderr, "0x%08x: WorkerDestroy\n", worker);
    if (worker->thread) {        
        // Wait for the worker to finish working
        WorkerWaitForCompletion(worker);
        
        // Then tell it to exit (this assumes that no other thread will enqueue more
        // work between this code and the call above).
        pthread_mutex_lock(&worker->mutex);
        worker->state = WorkerStoppedState;
        pthread_cond_signal(&worker->condition);

        while (worker->thread)
            pthread_cond_wait(&worker->condition, &worker->mutex);
        pthread_mutex_unlock(&worker->mutex);
    }
    
    pthread_cond_destroy(&worker->condition);
    pthread_mutex_destroy(&worker->mutex);
}


void WorkerStartBackgroundThread(Worker *worker)
{
    pthread_attr_t attr;
    
    //fprintf(stderr, "0x%08x: WorkerStartBackgroundThread\n", worker);
    pthread_mutex_lock(&worker->mutex);
    if (worker->thread) {
        //fprintf(stderr, "WorkerStartBackgroundThread -- there is already a worker thread for this work buffer (thread = 0x%08x\n", worker->thread);
        pthread_mutex_unlock(&worker->mutex);
        return;
    }

    worker->state = WorkerDoneState;
    
    pthread_attr_init(&attr);
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_create(&worker->thread, &attr, _WorkerThread, worker);
    pthread_attr_destroy(&attr);
    pthread_mutex_unlock(&worker->mutex);
}


void WorkerWaitForCompletion(Worker *worker)
{
    //fprintf(stderr, "0x%08x: WorkerWaitForCompletion\n", worker);
    pthread_mutex_lock(&worker->mutex);
    while (worker->state == WorkerBusyState) {
        //fprintf(stderr, "0x%08x:  WorkerWaitForCompletion: worker->state = %d\n", worker, worker->state);
        pthread_cond_wait(&worker->condition, &worker->mutex);
    }
    pthread_mutex_unlock(&worker->mutex);
}

void WorkerSignalMoreWork(Worker *worker)
{
    //fprintf(stderr, "0x%08x: WorkerSignalMoreWork\n", worker);
    
    // This assumes that only one thread is enqueueing work
    WorkerWaitForCompletion(worker);

    pthread_mutex_lock(&worker->mutex);
    worker->state = WorkerPendingState;
    pthread_cond_signal(&worker->condition);
    pthread_mutex_unlock(&worker->mutex);
}


static void *_WorkerThread(void *arg)
{
    Worker *worker = arg;
    
    while (1) {
        //fprintf(stderr, "0x%08x: _WorkerThread waiting for work\n",  worker);
        pthread_mutex_lock(&worker->mutex);
        while (worker->state == WorkerDoneState) {
            //fprintf(stderr, "0x%08x:  _WorkerThread: worker->state = %d\n", worker, worker->state);
            pthread_cond_wait(&worker->condition, &worker->mutex);
        }
        
        if (worker->state == WorkerPendingState) {
            //fprintf(stderr, "0x%08x:  _WorkerProcessWorkBuffer\n", worker);
            worker->function(worker);
            worker->state = WorkerDoneState;
            pthread_cond_signal(&worker->condition);
            pthread_mutex_unlock(&worker->mutex);
        } else if (worker->state == WorkerStoppedState) {
            //fprintf(stderr, "0x%08x:  Stopping\n", worker);
            worker->thread = NULL;
            pthread_cond_signal(&worker->condition);
            pthread_mutex_unlock(&worker->mutex);
            break;
        } else {
            //fprintf(stderr, "0x%08x: _WorkerThread got woken with state %d\n", worker, worker->state);
            pthread_mutex_unlock(&worker->mutex);
        }
    }
    
    return arg;
}

