// Copyright 2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

extern "C" {
#import "altivec.h"
#import "julia.h"
}

// I didn't bother to filter out all of my quaternion J set code when building
// this screen saver.  This particular screen saver really only does complex operations
// but it was easier to leave the quaternion class in place.
#import "quaternion.hxx"


void qJuliaComputeOrbit(Julia *julia)
{
    quaternion    *orbit, q, cl, qn;
    unsigned int   i;

    q = quaternion(0, 0, 0, 0);
    cl = quaternion(julia->c[0], julia->c[1], julia->c[2], julia->c[3]);
    orbit = (quaternion *)julia->orbit;

    for (i = 0; i < julia->N; i++) {
        float r2, i2, j2, k2;

        orbit[i] = q;
        //printf("  %03d: (%f,%f,%f,%f)\n", i, q.r, q.i, q.j, q.k);

        r2 = q.r * q.r;
        i2 = q.i * q.i;
        j2 = q.j * q.j;
        k2 = q.k * q.k;

        if (r2 + i2 + j2 + k2 > 4.0)
            break;

        qn.r = r2 - i2 - j2 - k2 + cl.r;
        qn.i = 2 * q.r * q.i + cl.i;
        qn.j = 2 * q.r * q.j + cl.j;
        qn.k = 2 * q.r * q.k + cl.k;

        q = qn;
    }

    julia->n = i;
}


#define EPSILON (0.00001)

static void _qCycleLength(Julia *julia)
{
    int           i;
    quaternion   *orbit, end;
    float         epsilonSquared;
    unsigned int  N;
    
    // Look backwards from the end of the orbit array for another point
    // that is within some epsilon.
    //printf("Looking for cycle:\n");

    orbit = (quaternion *)julia->orbit;
    N = julia->N;
    
    epsilonSquared = EPSILON * EPSILON;
    end = orbit[N - 1];
    for (i = N - 2; i >= 0; i--) {
        quaternion diff;
        float magSquared;

        diff = orbit[i] - end;
        magSquared = diff.magnitudeSquared();
        //printf("  %03d: (%f,%f,%f,%f) --> %f\n", i, diff.r, diff.i, diff.j, diff.k, magSquared);
        if (magSquared < epsilonSquared)
            break;
    }

    if (i != 0) {
        julia->cycleLength = N - i - 1;
        //printf("cycle length = %d\n", cycleLength);
    } else {
        julia->cycleLength = 0;
        //printf("no cycle found\n");
    }
}

// Computes the Mandelbrot orbit for the point defining this Julia set.
// This is necessary for estimating the distance from this point to
// the Mandlebrot set.  Points outside the M set are dust (totally
// disconnected, no area), points inside are totally connected.
// Points too far from the M set are unlikely to form 'interesting'
// Julia sets.
static void _qMandelbrotComputeOrbit(Julia *julia)
{
    quaternion    *orbit, q, c, qn;
    unsigned int   i;

    c = quaternion(julia->c[0], julia->c[1], julia->c[2], julia->c[3]);
    q = c;
    orbit = (quaternion *)julia->orbit;

    for (i = 0; i < julia->N; i++) {
        float r2, i2, j2, k2;

        orbit[i] = q;
        //printf("  %03d: (%f,%f,%f,%f)\n", i, q.r, q.i, q.j, q.k);

        r2 = q.r * q.r;
        i2 = q.i * q.i;
        j2 = q.j * q.j;
        k2 = q.k * q.k;

        if (r2 + i2 + j2 + k2 > 4.0)
            break;

        qn.r = r2 - i2 - j2 - k2 + c.r;
        qn.i = 2 * q.r * q.i + c.i;
        qn.j = 2 * q.r * q.j + c.j;
        qn.k = 2 * q.r * q.k + c.k;

        q = qn;
    }

    julia->n = i;
}


#define DISTANCE_OVERFLOW (1.0e200)
#define DISTANCE_DELTA    (0.1)      //???

// This computes the distance to the Mandelbrot set, given a
// Mandelbrot orbit (NOT a Julia orbit) for the point defining this
// Julia set.  Note that this algorithm is VERY similar to the Julia
// case above.  The only difference is in adding one in the loop.
static void _qMandelbrotDistanceEstimate(Julia *julia)
{
    unsigned long  k = 0;
    quaternion     dz(1.0, 0.0, 0.0, 0.0);
    double         a, mags;
    quaternion    *orbit;
    unsigned long  n;

    orbit = (quaternion *)julia->orbit;
    n = julia->n;

    mags = 1.0;

    while (k < n && mags < DISTANCE_OVERFLOW) {
        dz = (2.0 * (orbit[k] * dz)) + 1.0; // note the added one here, where the Julia version doesn't have it
        mags = dz.magnitudeSquared();
        k++;
    }
    if (mags >= DISTANCE_OVERFLOW) {
        julia->mandelbrotDistanceType = VeryCloseDistance;
        julia->mandelbrotDistanceEstimate = 0.0;
    }

    a = sqrt(orbit[n].magnitudeSquared());
    julia->mandelbrotDistanceEstimate = 0.5 * a * log(a) / sqrt(mags);

    if (julia->mandelbrotDistanceEstimate < DISTANCE_DELTA)
        julia->mandelbrotDistanceType = DeltaCloseDistance;
    else
        julia->mandelbrotDistanceType = NotCloseDistance;
}


void qJulia(Julia *julia)
{
    // First, compute the forward orbit
    qJuliaComputeOrbit(julia);

    // If we didn't escape, try to find a cycle
    if (julia->n == julia->N) {
        julia->mandelbrotDistanceType = InSetDistance;
        julia->mandelbrotDistanceEstimate = 0.0;
        _qCycleLength(julia);
    } else {
        // otherwise, try to estimate the distance from the point defining the set
        // to the Mandelbrot set.  To do this, we need to compute the Mandelbrot orbit.
        _qMandelbrotComputeOrbit(julia);

        if (julia->n == julia->N) {
            // Didn't escape, something weird is going on.  We should be outside
            // the M set!
            julia->mandelbrotDistanceType = InSetDistance;
            julia->mandelbrotDistanceEstimate = 0.0;
        } else {
            _qMandelbrotDistanceEstimate(julia);
        }
    }
}



// This structure should always be allocated quad word aligned.
typedef struct _cJuliaV4Args {
    float          zStart[2];
    float          c[2];
    float          xStep[2];
    float          yStep[2];
    unsigned long  maxI[4];
    unsigned long  positionX[4];
    unsigned long  positionY[4];
    unsigned int   pixelsWide, pixelsHigh;
    unsigned int   rowLength;
    unsigned int  *iterations;
} cJuliaV4Args;


// The bulk assembly Altivec computation function
extern "C" void vcComputeJulia4(cJuliaV4Args *args);

// The bulk C Altivec computation function
static void vcCComputeJulia4(cJuliaV4Args *args)
{
#ifdef __ALTIVEC__
    static vector bool long selectionMasks[4] = {
        (vector bool long)(0xffffffff, 0x00000000, 0x00000000, 0x00000000),
        (vector bool long)(0x00000000, 0xffffffff, 0x00000000, 0x00000000),
        (vector bool long)(0x00000000, 0x00000000, 0xffffffff, 0x00000000),
        (vector bool long)(0x00000000, 0x00000000, 0x00000000, 0xffffffff)
    };
    const vector unsigned char rotate[] = {
        (vector unsigned char)(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15),
        (vector unsigned char)(4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 0, 1, 2, 3),
        (vector unsigned char)(8, 9, 10, 11, 12, 13, 14, 15, 0, 1, 2, 3, 4, 5, 6, 7),
        (vector unsigned char)(12, 13, 14, 15, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11),
    };
    const vector float radius = (vector float)(4.0, 4.0, 4.0, 4.0);
    const vector float two = (vector float)(2.0, 2.0, 2.0, 2.0);
    vector unsigned one = (vector unsigned)(1, 1, 1, 1);
    vector float zero;
    vector unsigned vConvert;
    vector bool long doneMask;
    vector unsigned maxIter, iter;
    vector float zrNext, ziNext;
    vector float zrLeft, ziLeft;
    vector float cr, ci;
    vector float xrStep, xiStep, yrStep, yiStep;
    vector float zr, zi, zr2, zi2;
    vector float mags;
    unsigned int posX[4], posY[4];
    unsigned int width, height, rowLength, pixel;
    unsigned int xNext, yNext, x, y;
    unsigned int counter, tmp, tmp2, seloff;
    vector float vtmp1;
    vector bool long vtmp2, vtmp3, selMask;
    unsigned int *image;
    
    image = args->iterations;
    rowLength = args->rowLength;
    width = args->pixelsWide;
    height = args->pixelsHigh;
    
    zero = vec_xor(zero, zero);
    
    // Loads both z and c
    vtmp1 = vec_ld(0, args->zStart);
    
    zrLeft = zrNext = vec_splat(vtmp1, 0);
    ziLeft = ziNext = vec_splat(vtmp1, 1);
    cr = vec_splat(vtmp1, 2);
    ci = vec_splat(vtmp1, 3);
    
    // Loads both x and y step
    vtmp1 = vec_ld(0, args->xStep);
    
    xrStep = vec_splat(vtmp1, 0);
    xiStep = vec_splat(vtmp1, 1);
    yrStep = vec_splat(vtmp1, 2);
    yiStep = vec_splat(vtmp1, 3);
    
    maxIter = vec_ld(0, args->maxI);
    
    // Now we need to create the first four input points.  If we inline the code
    // to create them here, we will either have to do a lot of work to make sure
    // we deal with the width correctly (we might have a single pixel width).
    // Instead, we will use the point reloading code below.  To signal to it
    // that our points are empty (and no pixel should be written), we'll start
    // out with the ITER counter set to 0xffffffff for each point.
    iter = vec_xor(iter, iter);
    iter = vec_nor(iter, iter);
    doneMask = (vector bool long)iter;
    
    xNext = yNext = 0;
    
    goto InsertStart;
    
PointLoop:
IterationLoop:

    //
    // Compute an iteration of the Julia equation.  This is the most critical section
    // of the code.  The scheduling here must be very good.
    //
    vtmp1 = vec_madd(zr, zi, zero);
    zr2   = vec_madd(zr, zr, zero);
    zi2   = vec_madd(zi, zi, zero);
    iter  = vec_add(iter, one);
    zi    = vec_madd(vtmp1, two, ci);
    zr    = vec_sub(zr2, zi2);
    mags  = vec_add(zr2, zi2);
    zr    = vec_add(zr, cr);
    
    vtmp2    = vec_cmpeq(iter, maxIter);
    vtmp3    = vec_cmpgt(mags, radius);
    doneMask = vec_or(vtmp2, vtmp3);
    
    // If no points are done, then do another loop
    if (vec_all_eq(doneMask, (vector bool long)zero))
        goto IterationLoop;

InsertStart:
    counter = 4;
    seloff = 0;

InsertLoop:
    selMask = vec_ld(seloff, selectionMasks);
    vtmp2   = vec_and(doneMask, selMask);      // check if this point is zone

    // If the point is done, the resulting mask will be equal to the selMask
    // and zero otherwise
    if (vec_any_ne(vtmp2, selMask))
        // Wasn't equal, skip to the next point by jumping to the
        // bottom of the loop (which will handle decrementing the loop counter
        // and terminating the loop if we are done).
        goto InsertLoopBottom;

    // This position IS done.  Figure out the index of the finished result.  We can
    // compute this by shifting SELOFF down by 4 bits (dividing by 16).  But we
    // really want the index*sizeof(unsigned long), so we'll only shift by 2.
    tmp = seloff >> 2;
    
    // We might need to store the result.  If we are just starting, the iteration count
    // for this point will be 0xffffffff.  If this is the case, we should NOT store any
    // iteration count (since this slot hasn't been initialized yet).
    vtmp2 = (vector bool long)vec_and(selMask, iter);
    if (vec_all_eq(vtmp2, selMask))
        goto LoadNewPoint;

    // We need to store its results.  First, compute the address to
    // store the iteration counter in.  
    x = *(unsigned int *)((char *)posX + tmp);
    y = *(unsigned int *)((char *)posY + tmp);
    pixel = y * width + x;

    // We need to extract and store the integer iteration counter.  I don't grok the stvewx
    // instruction totally yet, so we'll rotate the slot we are interested in to the first
    // position, store it and load it back into an integer register.
    vtmp2 = (vector bool long)vec_ld(seloff, rotate);
    vtmp3 = (vector bool long)vec_perm((vector unsigned char)iter, (vector unsigned char)iter, (vector unsigned char)vtmp2);
    vec_ste((vector unsigned)vtmp3, 0, (unsigned long *)&vConvert);

    // We now want to write out the counter  to allow the higher level algorithm to do
    // block fills.  If I understood how stvewx really worked, I might be able to just store
    // the value in the right position in the first place, but memory bandwidth really isn't
    // the problem in this computation :)
    tmp2 = *(unsigned int *)&vConvert;
    image[pixel] = tmp2;
    
LoadNewPoint:

    // We might not have any more work to do.  This is the case if the next point
    // to enqueue is at (0, HEIGHT).
    if (xNext == 0 && yNext == height)
        goto StillValid;
        
    // No more pixels left to process.  Zero out any values in this point
    // and zero out the values used to update it during the iteration.
    // This will prevent it from ever being reported as done again, allowing
    // us to continue iterating until all pixels are done.
    zr = vec_sel(zr, zero, selMask);
    zi = vec_sel(zi, zero, selMask);
    cr = vec_sel(cr, zero, selMask);
    ci = vec_sel(ci, zero, selMask);
    one = vec_sel((vector unsigned)one, (vector unsigned)zero, selMask);

    // If ONEI is all zeros, then all the enqueued points have been flushed out
    // and we are *completely* done.
    if (vec_all_eq((vector unsigned)one, (vector unsigned)zero))
        goto Done;

    // Otherwise, continue looping through any other points that might need inserting
    goto InsertLoopBottom;
    
StillValid:
    // Now, we need to prepare this new slot with a new point and iteration counter
    zr = vec_sel(zr, zrNext, selMask);
    zi = vec_sel(zi, ziNext, selMask);
    zrNext = vec_add(zrNext, xrStep);
    ziNext = vec_add(ziNext, xiStep);
    
    // Zero out the iteration counter for this point
    iter = vec_sel((vector unsigned)iter, (vector unsigned)zero, selMask);
    
    // Store the indicies of this point
    *(unsigned int *)((char *)posX + tmp) = xNext;
    *(unsigned int *)((char *)posY + tmp) = yNext;

    // Update the indicies of the next point to be done
    xNext++;
    if (xNext != rowLength)
        goto InsertLoopBottom;
        
    // Ran off the line.  Need to step to the next.  If this was the last scanline, we're
    // really out of pixels to enqueue.  However, there are still potentially some pixels
    // left in the current registers that aren't done.  The issue is how to most easily
    // finish those pixels.  The solution we will use here is to zero out sections of key
    // registers used in performing the calculations.  This will ensure that these points
    // are never considered done and pixels will never be written for these points.  This
    // will allow the other points to complete.  Once all of the slots are zeroed out,
    // we know that we are really done.
    xNext = 0;
    yNext++;

    // Reset the next point position to the left edge
    zrNext = zrLeft = vec_sub(zrLeft, yrStep);
    ziNext = ziLeft = vec_sub(ziLeft, yiStep);

InsertLoopBottom:
    // Loop through the remaining points
    seloff += 16;
    counter--;
    if (!counter)
        goto InsertLoop;
    
    // Otherwise, we are done inserting points.  Resume processing    
    goto PointLoop;
    
Done:
    // Done.  Load our saved non-volatile registers and return
    return;
        
#endif
}




static void _cComputeJuliaImage4SubRect(cJuliaV4Args *args,
                                        float zStartR,
                                        float zStartI,
                                        unsigned int startX,
                                        unsigned int startY,
                                        unsigned int boxWidth,
                                        unsigned int boxHeight,
                                        unsigned int imageWidth,
                                        unsigned int imageHeight,
                                        unsigned int *iterations,
                                        int debug,
                                        int useC)
{
    if (debug) {
        /*
        fprintf(stderr, "   (%f, %f) (%d, %d, %d, %d) (%d, %d)\n",
                zStartR, zStartI,
                startX, startY,
                boxWidth, boxHeight,
                imageWidth, imageHeight);
         */
    }

    args->zStart[0] = zStartR + args->xStep[0] * startX - args->yStep[0] * startY;
    args->zStart[1] = zStartI + args->xStep[1] * startX - args->yStep[1] * startY;
    args->iterations = iterations + imageWidth * startY + startX;
    args->rowLength = boxWidth;
    args->pixelsWide = imageWidth;
    args->pixelsHigh = boxHeight;

    if (useC)
      vcCComputeJulia4(args);
    else
      vcComputeJulia4(args);
}


static void _cComputeJuliaImage4Recurse(cJuliaV4Args *args,
                                        float zStartR,
                                        float zStartI,
                                        unsigned int startX,
                                        unsigned int startY,
                                        unsigned int boxWidth,
                                        unsigned int boxHeight,
                                        unsigned int imageWidth,
                                        unsigned int imageHeight,
                                        unsigned int *iterations,
                                        unsigned int useC)
{
    unsigned int  borderCount;
    unsigned int *pixel;
    unsigned int  x, y;
    
    //
    // Check if the entire border is the same iteration count.
    //
    /*
    fprintf(stderr, "(%f, %f) (%d, %d, %d, %d) (%d, %d)\n",
            zStartR, zStartI,
            startX, startY,
            boxWidth, boxHeight, imageWidth, imageHeight);
     */
    // Get the count for the first pixel
    pixel = &iterations[startY * imageWidth + startX];
    borderCount = *pixel;

#warning The method below ends up checking the corner pixels twice each
    // Check the top
    for (x = 0; x < boxWidth; x++) {
        if (*pixel != borderCount)
            goto NoFill;
        pixel++;
    }

    // Check the right side
    pixel = &iterations[startY * imageWidth + startX + boxWidth - 1];
    for (y = 0; y < boxHeight; y++) {
        if (*pixel != borderCount)
            goto NoFill;
        pixel += imageWidth;
    }

    // Check the bottom
    pixel = &iterations[(startY + boxHeight - 1) * imageWidth + startX + boxWidth - 1];
    for (x = 0; x < boxWidth; x++) {
        if (*pixel != borderCount)
            goto NoFill;
        pixel--;
    }

    // Check the left side
    pixel = &iterations[(startY + boxHeight - 1) * imageWidth + startX];
    for (y = 0; y < boxHeight; y++) {
        if (*pixel != borderCount)
            goto NoFill;
        pixel -= imageWidth;
    }

    // Looks like we can fill it!  We'll only write to the interior (inset by one pixel on all sides)
    pixel = &iterations[(startY + 1) * imageWidth + startX + 1];
    for (y = 0; y < boxHeight - 2; y++) {
        for (x = 0; x < boxWidth - 2; x++) {
            *pixel = borderCount;
            pixel++;
        }
        pixel = pixel + imageWidth - (boxWidth - 2);
    }

    return;
    
NoFill:
        
    // If the width/height are too small, just do the whole block.  'Too small' here was
    // determined empirically by testing the execution time for the set (-.12, .74),
    // view width = 3.0, 256 iteration max, 640x480 pixels.  YMMV based on 'c' and the
    // aspect ratio.
    if ((boxWidth - 2) * (boxHeight - 2) < 64) {
        _cComputeJuliaImage4SubRect(args,
                                    zStartR, zStartI,
                                    startX + 1, startY + 1,
                                    boxWidth - 2, boxHeight - 2,
                                    imageWidth, imageHeight,
                                    iterations, 0, useC);
    } else {
        // Otherwise, split the box in half both horizontally and vertically and recurse
        unsigned int centerX, centerY;

        centerX = boxWidth/2;
        centerY = boxHeight/2;
        
        // down the center vertically
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    startX + centerX, startY + 1,
                                    1, boxHeight - 2,
                                    imageWidth, imageHeight,
                                    iterations, 1, useC);

        // across the center hoizontally
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    startX + 1, startY + centerY,
                                    boxWidth - 2, 1,
                                    imageWidth, imageHeight,
                                    iterations, 1, useC);

        // Recurse on the upper-left corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    startX, startY,
                                    centerX + 1, centerY + 1,
                                    imageWidth, imageHeight,
                                    iterations, useC);

        // Recurse on the upper-right corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    startX + centerX, startY,
                                    boxWidth - centerX, centerY + 1,
                                    imageWidth, imageHeight,
                                    iterations, useC);

        // Recurse on the lower-left corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    startX, startY + centerY,
                                    centerX + 1, boxHeight - centerY,
                                    imageWidth, imageHeight,
                                    iterations, useC);

        // Recurse on the lower-right corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    startX + centerX, startY + centerY,
                                    boxWidth - centerX, boxHeight - centerY,
                                    imageWidth, imageHeight,
                                    iterations, useC);
    }
}

// This higher level C version takes advantage of the fact that if a Julia set is connected,
// then any loop comprised of the same iteration count defines a region that is entirely of that
// same iteration count (except for the degenerate case where the loop goes around the whole set).
// We recursively descend, dividing up the area until all the pixels around a block are of the
// same count.  The region is then filled with that value.
void cComputeJuliaImage4(color_t *image,
                         const float center[2],
                         const float dx[2],
                         const float dy[2],
                         unsigned int pixelsWide,
                         unsigned int pixelsHigh,
                         const float c[2],
                         unsigned int maxIterations,
                         color_t clut[256],
                         int isConnected, int useC)
{
    cJuliaV4Args *args;
    unsigned int *iterations;
    unsigned int  pixelIndex, pixelCount;
    float zStartR, zStartI;
    
    vec_alloca(args, sizeof(*args));

    args->c[0] = c[0];
    args->c[1] = c[1];

    args->xStep[0] = dx[0];
    args->xStep[1] = dx[1];

    args->yStep[0] = dy[0];
    args->yStep[1] = dy[1];

    args->maxI[0] = maxIterations;
    args->maxI[1] = maxIterations;
    args->maxI[2] = maxIterations;
    args->maxI[3] = maxIterations;

    iterations = (unsigned int *)malloc(sizeof(unsigned int) * pixelsWide * pixelsHigh);    

    zStartR = center[0] - dx[0] * pixelsWide/2.0 + dy[0] * pixelsHigh/2.0;
    zStartI = center[1] - dx[1] * pixelsWide/2.0 + dy[1] * pixelsHigh/2.0;

    if (!isConnected) {
        args->zStart[0] = zStartR;
        args->zStart[1] = zStartI;

        args->iterations = iterations;
        args->rowLength = pixelsWide;
        args->pixelsWide = pixelsWide;
        args->pixelsHigh = pixelsHigh;

        if (useC)
            vcCComputeJulia4(args);
        else
            vcComputeJulia4(args);
    } else {
        // We always want to subdivide at least once to avoid falsely filling in the whole image
        // (since we are centered on the origin, we know that we'll cut through the set with this
        // subdivision.
        unsigned int centerX, centerY;

        centerX = pixelsWide/2;
        centerY = pixelsHigh/2;

        // top edge
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    0, 0,
                                    pixelsWide, 1,
                                    pixelsWide, pixelsHigh,
                                    iterations, 0, useC);

        // bottom edge
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    0, pixelsHigh - 1,
                                    pixelsWide, 1,
                                    pixelsWide, pixelsHigh,
                                    iterations, 0, useC);

        // left edge
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    0, 1,
                                    1, pixelsHigh - 2,
                                    pixelsWide, pixelsHigh,
                                    iterations, 0, useC);

        // right edge
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    pixelsWide - 1, 1,
                                    1, pixelsHigh - 2,
                                    pixelsWide, pixelsHigh,
                                    iterations, 0, useC);

        // down the center vertically
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    centerX, 1,
                                    1, pixelsHigh - 2,
                                    pixelsWide, pixelsHigh,
                                    iterations, 0, useC);

        // across the center hoizontally
        _cComputeJuliaImage4SubRect(args, zStartR, zStartI,
                                    1, centerY,
                                    pixelsWide - 2, 1,
                                    pixelsWide, pixelsHigh,
                                    iterations, 0, useC);


        // Recurse on the upper-left corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    0, 0,
                                    centerX + 1, centerY + 1,
                                    pixelsWide, pixelsHigh,
                                    iterations, useC);

        // Recurse on the upper-right corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    centerX, 0,
                                    pixelsWide - centerX, centerY + 1,
                                    pixelsWide, pixelsHigh,
                                    iterations, useC);

        // Recurse on the lower-left corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    0, centerY,
                                    centerX + 1, pixelsHigh - centerY,
                                    pixelsWide, pixelsHigh,
                                    iterations, useC);

        // Recurse on the lower-right corner
        _cComputeJuliaImage4Recurse(args, zStartR, zStartI,
                                    centerX, centerY,
                                    pixelsWide - centerX, pixelsHigh - centerY,
                                    pixelsWide, pixelsHigh,
                                    iterations, useC);
    }

    // Now, do the color lookups
    pixelCount = pixelsWide * pixelsHigh;
    for (pixelIndex = 0; pixelIndex < pixelCount; pixelIndex++) {
        image[pixelIndex] = clut[iterations[pixelIndex]];
    }
    
    free(iterations);
}

