// Copyright 2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.
//
// $Header: /Network/Users/Omni/bungi/Library/TimCVS/Tim/Altivec/Julia/OmniJulia/quaternion.hxx,v 1.2 2000/10/15 02:23:17 bungi Exp $

extern "C" {
#import <math.h>
#import <stdio.h>
}

class quaternion {

public:

    float r, i, j, k;

    inline quaternion() {
    };

    inline quaternion(float r0, float i0, float j0, float k0) {
        r = r0;
        i = i0;
        j = j0;
        k = k0;
    };

    inline quaternion &operator += (quaternion const &a) {
        r += a.r;
        i += a.i;
        j += a.j;
        k += a.k;

        return *this;
    };

    inline quaternion &operator -= (quaternion const &a) {
        r -= a.r;
        i -= a.i;
        j -= a.j;
        k -= a.k;

        return *this;
    };

    inline quaternion &operator *= (float const &a) {
        r *= a;
        i *= a;
        j *= a;
        k *= a;

        return *this;
    };

    inline quaternion &operator *= (quaternion const &a) {
        *this = *this * a;
        return *this;
    };

    inline quaternion &operator /= (float const &a) {
        float invA = (1.0 / a);

        // multiplies are faster than divides -- we might lose some precision, though
        r *= invA;
        i *= invA;
        j *= invA;
        k *= invA;

        return *this;
    };

    inline quaternion operator = (quaternion const &a) {
        r = a.r;
        i = a.i;
        j = a.j;
        k = a.k;

        return *this;
    };

    inline quaternion operator + (quaternion const &a) const {
        quaternion temp;

        temp.r = r + a.r;
        temp.i = i + a.i;
        temp.j = j + a.j;
        temp.k = k + a.k;

        return temp;
    };

    inline quaternion operator + (float const &a) const {
        quaternion temp;

        temp.r = r + a;
        temp.i = i;
        temp.j = j;
        temp.k = k;

        return temp;
    };

    inline quaternion operator - (quaternion const &a) const {
        quaternion temp;

        temp.r = r - a.r;
        temp.i = i - a.i;
        temp.j = j - a.j;
        temp.k = k - a.k;

        return temp;
    };

    inline quaternion operator - () const {
        quaternion temp;

        temp.r = -r;
        temp.i = -i;
        temp.j = -j;
        temp.k = -k;

        return temp;
    };

    // Declare this as a friend function to make sure we get a*b instead of b*a
    // since this is non-commutative
    inline friend quaternion operator *(const quaternion &a, const quaternion &b) {
        quaternion temp;

        temp.r = a.r * b.r - a.i * b.i - a.j * b.j - a.k * b.k;
        temp.i = a.r * b.i + a.i * b.r + a.j * b.k - a.k * b.j;
        temp.j = a.r * b.j - a.i * b.k + a.j * b.r + a.k * b.i;
        temp.k = a.r * b.k + a.i * b.j - a.j * b.i + a.k * b.r;

        return temp;
    };

    inline friend quaternion operator *(const quaternion &a, const float &b) {
        quaternion temp;

        temp.r = a.r * b;
        temp.i = a.i * b;
        temp.j = a.j * b;
        temp.k = a.k * b;

        return temp;
    }

    inline friend quaternion operator *(const float &b, const quaternion &a) {
        return a * b;
    }

    inline float dot (quaternion const &a) const {
        return r * a.r + i * a.i + j * a.j + k * a.k;
    };

    inline float magnitudeSquared() const {
        return this->dot(*this);
    };

    inline quaternion inverse() const {
        quaternion temp;
        float invMags;

        invMags = 1.0 / this->magnitudeSquared();

        temp.r =  r * invMags;
        temp.i = -i * invMags;
        temp.j = -j * invMags;
        temp.k = -k * invMags;

        return temp;
    };


    // Declare this as a friend function to make sure we get a/b instead of b/a
    // since this is non-commutative
    inline friend quaternion operator /(const quaternion &a, const quaternion &b) {
        return a * b.inverse();
    };

    inline quaternion operator / (float const &a) const {
        quaternion temp;
        float invA = (1.0 / a);

        // multiplies are faster than divides -- we might lose some precision, though
        temp.r = r * invA;
        temp.i = i * invA;
        temp.j = j * invA;
        temp.k = k * invA;

        return temp;
    };

    inline float distanceSquared(quaternion const &a) const {
        quaternion delta;

        delta = *this - a;
        return delta.magnitudeSquared();
    };

    inline bool withinEpsilon(quaternion const &a, float const &epsilon) const {
        float dr, di, dj, dk;

        if (((dr = fabs(r - a.r)) > epsilon) ||
            ((di = fabs(i - a.i)) > epsilon) ||
            ((dj = fabs(j - a.j)) > epsilon) ||
            ((dk = fabs(k - a.k)) > epsilon))
            return false;

        return (dr * dr + di * di + dj * dj + dk * dk) < epsilon;
    };

    inline quaternion normalized() const {
        quaternion temp;
        float             mag;

        mag = magnitudeSquared();

        if (mag > 0.0)
            temp = *this / sqrt(mag);
        else
            temp = quaternion(0.0, 0.0, 0.0, 0.0);

        return temp;
    };

    inline quaternion conjugate() const {
        quaternion temp;

        temp.r = r;
        temp.i = -i;
        temp.j = -j;
        temp.k = -k;

        return temp;
    }

    inline void print(const char *s) const {
        fprintf(stderr, "%s <%f, %f, %f, %f>\n", s, r, i, j, k);
    }
};
